__all__ = [
    "AdsClient",
    "AdsConnection",
    "AdsDatatype",
    "AdsDevice",
    "AdsException",
    "AdsState",
    "amspacket",
    "AmsPacket",
    "BinaryParser",
    "SymbolInfo",
    "HexBlock",
    "AdsIO",
    "ProcImage"
]

from pyads.symbolinfo import *
from pyads.adsdatatype import *
from pyads.adsexception import *
from pyads.adsstate import *
from pyads.amspacket import *
from pyads.binaryparser import *
from pyads.adsconnection import *
from pyads.adsclient import *
from pyads.adsdevice import *
from pyads.adsio import *
from pyads.procimage import *


def HexBlock(data, width = 8):
    i, result, currentHexLine, currentChrLine = 0, '', '', ''

    for byte in data:

        # next line, if required
        if (i == width):
            result += '%s %s\n' % (currentHexLine, currentChrLine)
            currentHexLine = ''
            currentChrLine = ''
            i = 0

        # python2 / python3 - normalize to numeric byte
        char = ord(byte) if isinstance(byte, str) else byte

        # append to lines
        currentHexLine += '%02x ' % char
        currentChrLine += '.' if (char < 32 or char > 126) else chr(char)
        i += 1

    # append last line
    result += '%s %s' % (currentHexLine, currentChrLine)
    return result


from .. import AmsPacket
from .. import AdsException

class AdsCommand:

    def __init__(self):
        pass


    CommandID = 0


    def CreateRequest(self):
        raise NotImplementedError()


    def CreateResponse(self, responsePacket):
        raise NotImplementedError()


    def GetAmsPacket(self, adsConnection):
        packet = AmsPacket(adsConnection)
        packet.CommandID = self.CommandID
        packet.StateFlags = 0x0004
        packet.Data = self.CreateRequest()

        return packet


    def Execute(self, adsClient):

        # create packet
        packet = self.GetAmsPacket(adsClient.AdsConnection)

        # send to client
        responsePacket = adsClient.SendAndRecv(packet)

        # check for error
        if (responsePacket.ErrorCode > 0):
            raise AdsException(responsePacket.ErrorCode)

        # return response object
        result = self.CreateResponse(responsePacket)
        if (result.Error > 0):
            raise AdsException(result.Error)

        return result







import struct
from .. import AmsPacket
from .adsresponse import AdsResponse

class ReadWriteResponse(AdsResponse):

    def __init__(self, responseAmsPacket):
        AdsResponse.__init__(self, responseAmsPacket)

        self.Length = struct.unpack_from('I', responseAmsPacket.Data, 4)[0]
        self.Data = responseAmsPacket.Data[8:]


    Length = 0

    Data = b'0'


    def __str__(self):
        result = "AdsReadWriteResponse:\n"
        result += AmsPacket.GetHexStringBlock(self.Data)
        return result




class ProcImage:


    def __init__(self, indexGroup, indexOffset, length):
        self.IndexGroup = indexGroup
        self.IndexOffset = indexOffset
        self.Length = length


    IndexGroup = 0

    IndexOffset = 0

    Length = 0

    ByteBuffer = None


    def Read(self, adsClient):
        readResult = adsClient.Read(self.IndexGroup, self.IndexOffset, self.Length)
        self.ByteBuffer = readResult.CreateBuffer()


    def Write(self, adsClient):
        writeData = self.ByteBuffer.raw
        adsClient.Write(self.IndexGroup, self.IndexOffset, writeData)



import struct
from .adsresponse import AdsResponse

class ReadStateResponse(AdsResponse):

    def __init__(self, responseAmsPacket):
        AdsResponse.__init__(self, responseAmsPacket)

        self.AdsState = struct.unpack_from('H', responseAmsPacket.Data, 4)[0]
        self.DeviceState = struct.unpack_from('H', responseAmsPacket.Data, 6)[0]


    AdsState = 0

    DeviceState = 0


    def __str__(self):
        return "Ads/Device State: %s/%s" % (self.AdsState, self.DeviceState)




import re

class AdsConnection:

    def __init__(self, targetAms='0.0.0.0.0.0:0', sourceAms=None, targetIP=None):

        targetAmsInfo = self.ParseAmsIdPort(targetAms)
        if targetIP is None:
            self.TargetIP = targetAmsInfo[0]
        else:
            self.TargetIP = targetIP
        self.TargetAmsID = targetAmsInfo[1]
        self.TargetAmsPort = targetAmsInfo[2]

        if sourceAms == None:
            sourceAms = self.GetDefaultSourceAms()

        sourceAmsInfo = self.ParseAmsIdPort(sourceAms)
        self.SourceAmsID = sourceAmsInfo[1]
        self.SourceAmsPort = sourceAmsInfo[2]



    def ParseAmsIdPort(self, amsconnection):
        pattern = '([0-9\.]+):([0-9]+){1}'
        match = re.match(pattern, amsconnection)
        if (match == None):
            raise Exception("amsid port format not valid. You must enter i.e. '192.168.1.17.1.1:800'.")

        amsID = str(match.group(1))
        tcpIP = '.'.join(x for x in amsID.split('.')[:4])
        amsPort = int(match.group(2))

        return (tcpIP, amsID, amsPort)



    TargetIP = ''
    TargetAmsID = ''
    TargetAmsPort = 0

    SourceAmsID = ''
    SourceAmsPort = 0


    def GetDefaultSourceAms(self):
        return '0.0.0.0.0.0:32905'


    def __str__(self):
        return "%s:%s --> %s:%s" % (
            self.SourceAmsID,
            self.SourceAmsPort,
            self.TargetAmsID,
            self.TargetAmsPort,
        )



import struct
from .. import AdsException

class AdsResponse:

    def __init__(self, responseAmsPacket):
        self.Error = struct.unpack_from('I', responseAmsPacket.Data)[0]

        if (self.Error > 0):
            raise AdsException(self.Error)


    Error = 0



__all__ = [
    "AdsCommand",
    "AdsResponse",
    "DeviceInfoCommand",
    "DeviceInfoResponse",
    "ReadCommand",
    "ReadResponse",
    "ReadStateCommand",
    "ReadStateResponse",
    "ReadWriteCommand",
    "ReadWriteResponse",
    "WriteCommand",
    "WriteResponse",
    "WriteControlCommand",
    "WriteControlResponse"
]

from .adscommand import *
from .adsresponse import *
from .deviceinfocommand import *
from .deviceinforesponse import *
from .readcommand import *
from .readresponse import *
from .readstatecommand import *
from .readstateresponse import *
from .readwritecommand import *
from .readwriteresponse import *
from .writecommand import *
from .writeresponse import *
from .writecontrolcommand import *
from .writecontrolresponse import *



from .binaryparser import BinaryParser
import pyads

class AmsPacket():

    def __init__(self, connection = None):
        if(connection != None):
            self.TargetAmsID = connection.TargetAmsID
            self.TargetAmsPort = connection.TargetAmsPort
            self.SourceAmsID = connection.SourceAmsID
            self.SourceAmsPort = connection.SourceAmsPort



    TargetAmsID = ''
    """the ams-net-id of the destination device (6 bytes)"""

    TargetAmsPort = 0
    """the ams-port to use (2 bytes, UInt16)"""

    SourceAmsID = '0'
    """the ams-net-id of the sender device (6 bytes)"""

    SourceAmsPort = 0
    """the ams-port of the sender (2 bytes, UInt16)"""

    CommandID = 0
    """command-id (2 bytes, UInt16)"""

    StateFlags = 0
    """state flags, i.e. 0x0004 for request. (2 bytes, UInt16)"""

    Length = 0
    """length of data (4 bytes, UInt32)"""

    ErrorCode = 0
    """error code of ads-response (4 bytes, UInt32)"""

    InvokeID = 0
    """free choosable number to identify request<->response  (4 bytes, UInt32)"""

    Data = b''
    """the ads-data to transmit"""


    @staticmethod
    def AmsNetIDToBytes(pointDottedBytes):
        return list(map(int, pointDottedBytes.split('.')))



    @staticmethod
    def AmsNetIDFromBytes(byteList):
        words = []

        for bt in byteList:
            words.append("%s" % ord(bt))

        return ".".join(words)



    def GetBinaryData(self):
        binary = BinaryParser()

        # tcp/ip header
        #binary.WriteUInt16(0)
        #binary.WriteUInt32(32 + len(self.Data))

        # amsheader
        # ams-target id & port
        binary.WriteBytes(AmsPacket.AmsNetIDToBytes(self.TargetAmsID))
        binary.WriteUInt16(self.TargetAmsPort)

        # ams-source id & port
        binary.WriteBytes(AmsPacket.AmsNetIDToBytes(self.SourceAmsID))
        binary.WriteUInt16(self.SourceAmsPort)

        # command id, state flags & data length
        binary.WriteUInt16(self.CommandID)
        binary.WriteUInt16(self.StateFlags)
        binary.WriteUInt32(len(self.Data))

        # error code & invoke id
        binary.WriteUInt32(self.ErrorCode)
        binary.WriteUInt32(self.InvokeID)

        # last but not least - the data
        binary.WriteBytes(self.Data)

        # return byte buffer
        return binary.ByteData


    @staticmethod
    def FromBinaryData(data = ''):
        binary = BinaryParser(data)
        result = AmsPacket()

        # amsheader
        # ams-target id & port
        result.TargetAmsID = AmsPacket.AmsNetIDFromBytes(binary.ReadBytes(6))
        result.TargetAmsPort = binary.ReadUInt16()

        # ams-source id & port
        result.SourceAmsID = AmsPacket.AmsNetIDFromBytes(binary.ReadBytes(6))
        result.SourceAmsPort = binary.ReadUInt16()

        # command id, state flags & data length
        result.CommandID = binary.ReadUInt16()
        result.StateFlags = binary.ReadUInt16()
        result.Length = binary.ReadUInt32()

        # error code & invoke id
        result.ErrorCode = binary.ReadUInt32()
        result.InvokeID = binary.ReadUInt32()

        # last but not least - the data
        result.Data = binary.ByteData[32:]

        # ready, give out!
        return result




    def __str__(self):
        result = "FROM %s:%s --> " % (self.SourceAmsID, self.SourceAmsPort)
        result += "%s:%s\n" % (self.TargetAmsID, self.TargetAmsPort)
        result += "Command ID:  %s\n" % self.CommandID
        result += "Invoke ID:   %s\n" % self.InvokeID
        result += "State Flags: %s\n" % self.StateFlags
        result += "Data Length: %s\n" % self.Length
        result += "Error:       %s\n" % self.ErrorCode

        if (len(self.Data) == 0):
            result += "Packet contains no data.\n"
        else:
            result += "Data:\n%s\n" % pyads.HexBlock(self.Data)

        return result



from .adsdatatype import AdsDatatype

class SymbolInfo:


    def __init__(self, name, indexGroup, indexOffset, adsDatatype, bitOffset = 0):
        self.Name = name
        self.IndexGroup = indexGroup
        self.IndexOffset = indexOffset
        self.AdsDatatype = adsDatatype
        self.BitOffset = bitOffset


    Name = ''

    IndexGroup = 0

    IndexOffset = 0

    BitOffset = 0

    AdsDatatype = AdsDatatype.Custom

    Value = None


    def WriteTo(self, byteBuffer):

        # byte shift needed, if bool!
        if (self.AdsDatatype == AdsDatatype.Bool):
            currentByte = AdsDatatype.UnpackFrom(AdsDatatype.UInt8, byteBuffer, self.IndexOffset)
            if (self.Value):
                newByte = currentByte | (1 << self.BitOffset)
            else:
                newByte = currentByte & ~(1 << self.BitOffset) & 0xF

            AdsDatatype.PackInto(AdsDatatype.UInt8, byteBuffer, self.IndexOffset, newByte)

        else:
            AdsDatatype.PackInto(self.AdsDatatype, byteBuffer, self.IndexOffset, self.Value)



    def ReadFrom(self, byteBuffer):

        if (self.AdsDatatype == AdsDatatype.Bool):
            result = AdsDatatype.UnpackFrom(AdsDatatype.UInt8, byteBuffer, self.IndexOffset)
            result = ((result & (1 << self.BitOffset)) == True)
        else:
            result = AdsDatatype.UnpackFrom(self.AdsDatatype, byteBuffer, self.IndexOffset)

        self.Value = result
        return result



    def __str__(self):
        return "%s [%s] (%08x, %08x%s)" % (
            self.Name,
            AdsDatatype.GetName(self.AdsDatatype),
            self.IndexGroup,
            self.IndexOffset,
            (".%s" % self.BitOffset) if self.AdsDatatype == AdsDatatype.Bool else ''
        )


from .adscommand import AdsCommand
from .readstateresponse import ReadStateResponse

class ReadStateCommand(AdsCommand):

    def __init__(self):
        self.CommandID = 0x0004


    def CreateRequest(self):
        return b''


    def CreateResponse(self, responsePacket):
        return ReadStateResponse(responsePacket)



import time
import select
import socket
import struct
import threading
import errno

from .amspacket import AmsPacket
from .adsconnection import AdsConnection
from .adsexception import AdsException
from .commands import *

class AdsClient:

    def __init__(self, adsConnection = None, amsTarget = None, amsSource = None, targetIP = None):
        if adsConnection != None and amsTarget == None and amsSource == None:
            self.AdsConnection = adsConnection

        elif amsTarget != None and adsConnection == None:
            self.AdsConnection = AdsConnection(amsTarget, amsSource, targetIP)

        else:
            raise Exception('You must specify either connection or adsTarget, not both.')

    MAX_RETRY_ON_FAIL = 3

    Debug = False

    RetryOnFail = 0

    AdsConnection = None

    AdsPortDefault = 0xBF02

    AdsIndexGroupIn = 0xF020

    AdsIndexGroupOut = 0xF030

    AdsChunkSizeDefault = 1024

    Socket = None

    _CurrentInvokeID = 0x8000

    _CurrentPacket = None


    @property
    def IsConnected(self):
        return self.Socket != None


    def Close(self):
        if (self.Socket != None):
            self.Socket.close()
            self.Socket = None



    def Connect(self):
        self.Close()
        self.Socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.Socket.settimeout(2)

        try:
            self.Socket.connect((self.AdsConnection.TargetIP, self.AdsPortDefault))
            self._BeginAsyncRead()
        except socket.error:
            raise AdsException(0x274c)



    def _BeginAsyncRead(self):
        self._AsyncReadThread = threading.Thread(target=self._AsyncRead)
        self._AsyncReadThread.start()


    def _AsyncRead(self):

        while self.IsConnected:
            try:
                ready = select.select([self.Socket], [], [], 0.1)

                if ready[0] and self.IsConnected:
                        newPacket = self.ReadAmsPacketFromSocket()
                        if (newPacket.InvokeID == self._CurrentInvokeID):
                            self._CurrentPacket = newPacket
                        else:
                            print("Packet dropped:")
                            print(newPacket)
            except (socket.error, select.error):
                self.Close()
                break



    def ReadAmsPacketFromSocket(self):

        # read default buffer
        response = self.Socket.recv(self.AdsChunkSizeDefault)

        # ensure correct beckhoff tcp header
        if(len(response) < 6):
            return None

        # first two bits must be 0
        if (response[0:2] != b'\x00\x00'):
            return None

        # read whole data length
        dataLen = struct.unpack('I', response[2:6])[0] + 6

        # read rest of data, if any
        while (len(response) < dataLen):
            nextReadLen = min(self.AdsChunkSizeDefault, dataLen - len(response))
            response += self.Socket.recv(nextReadLen)

        # cut off tcp-header and return response amspacket
        return AmsPacket.FromBinaryData(response[6:])


    def GetTcpHeader(self, amsData):
        # pack 2 bytes (reserved) and 4 bytes (length)
        # format _must_ be little endian!
        return struct.pack('<HI', 0, len(amsData))


    def SendAndRecv(self, amspacket):

        if not self.IsConnected:
            self.Connect()

        # prepare packet with invoke id
        self.PrepareCommandInvoke(amspacket)

        # send tcp-header and ams-data
        try:
            self.Socket.send(self.GetTCPPacket(amspacket))
        except socket.error as e:
            # if i fail Socket.send i try again for 3 times
            if self.RetryOnFail < self.MAX_RETRY_ON_FAIL:
                self.RetryOnFail += 1
                # if i have a BROKEN PIPE error i reconnect
                # the socket before try again
                if e.errno == errno.EPIPE:
                    self.Connect()
                return self.SendAndRecv(amspacket)
            else:
                self.RetryOnFail = 0
                raise AdsException(0x274c)

        # here's your packet
        return self.AwaitCommandInvoke()



    def GetTCPPacket(self, amspacket):

        # get ams-data and generate tcp-header
        amsData = amspacket.GetBinaryData()
        tcpHeader = self.GetTcpHeader(amsData)

        return tcpHeader + amsData



    def PrepareCommandInvoke(self, amspacket):
        if(self._CurrentInvokeID < 0xFFFF):
            self._CurrentInvokeID += 1
        else:
            self._CurrentInvokeID = 0x8000

        self._CurrentPacket = None
        amspacket.InvokeID = self._CurrentInvokeID

        if self.Debug:
            print(">>> sending ams-packet:")
            print(amspacket)



    def AwaitCommandInvoke(self):
        # unfortunately threading.event is slower than this oldschool poll :-(
        timeout = 0
        while (self._CurrentPacket == None):
            timeout += 0.001
            time.sleep(0.001)
            if (timeout > 10):
                raise AdsException(0x745)

        if self.Debug:
            print("<<< received ams-packet:")
            print(self._CurrentPacket)

        return self._CurrentPacket




    def ReadDeviceInfo(self):
        return DeviceInfoCommand().Execute(self)


    def Read(self, indexGroup, indexOffset, length):
        return ReadCommand(indexGroup, indexOffset, length).Execute(self)


    def Write(self, indexGroup, indexOffset, data):
        return WriteCommand(indexGroup, indexOffset, data).Execute(self)


    def ReadState(self):
        return ReadStateCommand().Execute(self)


    def WriteControl(self, adsState, deviceState, data = b''):
        return WriteControlCommand(adsState, deviceState, data).Execute(self)


    def AddDeviceNotification(self):
        raise NotImplementedError()


    def DeleteDeviceNotification(self):
        raise NotImplementedError()


    def ReadWrite(self, indexGroup, indexOffset, readLen, dataToWrite = b''):
        return ReadWriteCommand(indexGroup, indexOffset, readLen, dataToWrite).Execute(self)


    def __enter__(self):
        return self


    def __exit__(self, vtype, value, traceback):
        self.Close()




import struct
from .adscommand import AdsCommand
from .writecontrolresponse import WriteControlResponse

class WriteControlCommand(AdsCommand):

    def __init__(self, adsState, deviceState, data = b''):
        self.CommandID = 0x0005
        self.AdsState = adsState
        self.DeviceState = deviceState
        self.Data = data


    AdsState = 0
    DeviceState = 0
    Data = b''


    def CreateRequest(self):
        result = struct.pack('<HHI', self.AdsState, self.DeviceState, len(self.Data))
        result += self.Data
        return result


    def CreateResponse(self, responsePacket):
        return WriteControlResponse(responsePacket)



import copy
import struct
from .adsclient import AdsClient
from .adsdatatype import AdsDatatype
from .adsconnection import AdsConnection
from .binaryparser import BinaryParser

class AdsDevice(AdsClient):

    def __init__(self, adsConnection = None, amsTarget = None, amsSource = None, targetIP = None):
        AdsClient.__init__(self, adsConnection, amsTarget, amsSource, targetIP)


    def GetSymbolHandle(self, variableName):
        symbolData = self.ReadWrite(0xF003, 0x0000, 4, variableName.encode('ascii') + b'\x00').Data
        symbolHandle = struct.unpack("I", symbolData)[0]
        return symbolHandle


    def ReadByName(self, variableName, adsDatatype, length = None):
        symbolHandle = self.GetSymbolHandle(variableName)
        return self.ReadByHandle(symbolHandle, adsDatatype, length)


    def ReadByHandle(self, symbolHandle, adsDatatype, length = None):
        if length is None:
            length = AdsDatatype.GetSize(adsDatatype)
        data = self.Read(0xF005, symbolHandle, length).Data
        return AdsDatatype.Unpack(data, adsDatatype)


    def WriteByName(self, variableName, adsDatatype, value):
        symbolHandle = self.GetSymbolHandle(variableName)
        self.WriteByHandle(symbolHandle, adsDatatype, value)


    def WriteByHandle(self, symbolHandle, adsDatatype, value):
        valueRaw = AdsDatatype.Pack(value, adsDatatype)
        self.Write(0xF005, symbolHandle, valueRaw)




class AdsState():

    def __init__(self):
        pass

    Invalid = 0
    Idle = 1
    Reset = 2
    Init = 3
    Start = 4
    Run = 5
    Stop = 6
    SaveConfig = 7
    LoadConfig = 8
    PowerFailure = 9
    PowerGood = 10
    Error = 11
    Shutdown = 12
    Suspend = 13
    Resume = 14
    Config = 15
    ReConfig = 16




from .procimage import ProcImage
from .adsclient import AdsClient

class AdsIO:


    def __init__(self, client):
        self.Client = client


    Client = None
    ProcImageIn = ProcImage(AdsClient.AdsIndexGroupIn, 0x0000, 512)
    ProcImageOut = ProcImage(AdsClient.AdsIndexGroupOut, 0x0000, 512)

    _Variables = {}


    def Register(self, symbolInfo):
        self._Variables[symbolInfo.Name] = symbolInfo


    def Get(self, symbolName):
        return self._Variables[symbolName].Value


    def Set(self, symbolName, value):
        self._Variables[symbolName].Value = value


    def Initialize(self):
        self.ProcImageOut.Read(self.Client)
        self.ReadVariables(self.ProcImageOut)

        self.ProcImageIn.Read(self.Client)
        self.ReadVariables(self.ProcImageIn)


    def ReadAll(self):
        self.ProcImageIn.Read(self.Client)
        self.ReadVariables(self.ProcImageIn)


    def ReadVariables(self, procImage):
        for si in list(self._Variables.values()):
            if si.IndexGroup == procImage.IndexGroup:
                si.ReadFrom(procImage.ByteBuffer)


    def WriteAll(self):
        self.WriteVariables(self.ProcImageOut)
        self.ProcImageOut.Write(self.Client)


    def WriteVariables(self, procImage):
        for si in list(self._Variables.values()):
            if si.IndexGroup == procImage.IndexGroup:
                si.WriteTo(procImage.ByteBuffer)




import struct
from .adscommand import AdsCommand
from .writeresponse import WriteResponse

class WriteCommand(AdsCommand):

    def __init__(self, indexGroup, indexOffset, data):
        self.CommandID = 0x0003
        self.IndexGroup = indexGroup
        self.IndexOffset = indexOffset
        self.Data = data



    IndexGroup = 0

    IndexOffset = 0

    Data = b''


    def CreateRequest(self):
        result = struct.pack('<III', self.IndexGroup, self.IndexOffset, len(self.Data))
        result += self.Data
        return result


    def CreateResponse(self, responsePacket):
        return WriteResponse(responsePacket)



import struct


class AdsDatatype:

    Custom = 0x00
    Int8 = 0x01
    UInt8 = 0x02
    Int16 = 0x03
    UInt16 = 0x04
    Int32 = 0x05
    UInt32 = 0x06
    Int64 = 0x07
    UInt64 = 0x08
    Float = 0x09
    Double = 0x0A
    Bool = 0x0B
    String = 0x0C


    __infomapping__ = {
        Int8:   ('b', 1, 'Int8'),
        UInt8:  ('B', 1, 'UInt8'),
        Int16:  ('h', 2, 'Int16'),
        UInt16: ('H', 2, 'UInt16'),
        Int32:  ('i', 4, 'Int32'),
        UInt32: ('I', 4, 'UInt32'),
        Int64:  ('q', 8, 'Int64'),
        UInt64: ('Q', 8, 'UInt64'),
        Float:  ('f', 4, 'Float'),
        Double: ('d', 8, 'Double'),
        Bool:   ('?', 1, 'Bool'),
        String: (None, 80, 'String')
    }



    @staticmethod
    def GetPackFormat(adsDatatype):
        return AdsDatatype.__infomapping__.get(adsDatatype)[0]


    @staticmethod
    def GetSize(adsDatatype):
        return AdsDatatype.__infomapping__.get(adsDatatype)[1]


    @staticmethod
    def GetName(adsDatatype):
        return AdsDatatype.__infomapping__.get(adsDatatype)[2]


    @staticmethod
    def Pack(value, adsDatatype):

        packFmt = AdsDatatype.GetPackFormat(adsDatatype)

        if(packFmt != None):
            return struct.pack(packFmt, value)
        else:
            return value


    @staticmethod
    def PackInto(adsDatatype, byteBuffer, offset, value):

        packFmt = AdsDatatype.GetPackFormat(adsDatatype)

        if(packFmt == None):
            raise Exception("no pack format found.")

        struct.pack_into(packFmt, byteBuffer, offset, value)


    @staticmethod
    def UnpackFrom(adsDatatype, byteBuffer, offset):

        packFmt = AdsDatatype.GetPackFormat(adsDatatype)

        if(packFmt == None):
            raise Exception("no pack format found.")

        return struct.unpack_from(packFmt, byteBuffer, offset)[0]


    @staticmethod
    def Unpack(value, adsDatatype):

        packFmt = AdsDatatype.GetPackFormat(adsDatatype)

        if(packFmt != None):
            return struct.unpack(packFmt, value)[0]
        else:
            return value




class AdsException(Exception):


    def __init__(self, code):
        self.code = code


    def __str__(self):
        if(self.code in self.AdsCodeNumbers):
            return repr(self.AdsCodeNumbers[self.code])
        else:
            return repr("Error Code #%s" % self.code)


    AdsCodeNumbers = {
        0x1: "Internal error",
        0x2: "No Runtime",
        0x3: "Allocation locked memory error",
        0x4: "Insert mailbox error. Reduce the number of ADS calls",
        0x5: "Wrong receive HMSG",
        0x6: "target port not found",
        0x7: "target machine not found",
        0x8: "Unknown command ID",
        0x9: "Bad task ID",
        0xA: "No IO",
        0xB: "Unknown ADS command",
        0xC: "Win 32 error",
        0xD: "Port not connected",
        0xE: "Invalid ADS length",
        0xF: "Invalid ADS Net ID",
        0x10: "Low Installation level",
        0x11: "No debug available",
        0x12: "Port disabled",
        0x13: "Port already connected",
        0x14: "ADS Sync Win32 error",
        0x15: "ADS Sync Timeout",
        0x16: "ADS Sync AMS error",
        0x17: "ADS Sync no index map",
        0x18: "Invalid ADS port",
        0x19: "No memory",
        0x1A: "TCP send error",
        0x1B: "Host unreachable",
        0x700: "error class <device error>",
        0x701: "Service is not supported by server",
        0x702: "invalid index group",
        0x703: "invalid index offset",
        0x704: "reading/writing not permitted",
        0x705: "parameter size not correct",
        0x706: "invalid parameter value(s)",
        0x707: "device is not in a ready state",
        0x708: "device is busy",
        0x709: "invalid context (must be in Windows)",
        0x70A: "out of memory",
        0x70B: "invalid parameter value(s)",
        0x70C: "not found (files, ...)",
        0x70D: "syntax error in command or file",
        0x70E: "objects do not match",
        0x70F: "object already exists",
        0x710: "symbol not found",
        0x711: "symbol version invalid",
        0x712: "server is in invalid state",
        0x713: "AdsTransMode not supported",
        0x714: "Notification handle is invalid",
        0x715: "Notification client not registered",
        0x716: "no more notification handles",
        0x717: "size for watch too big",
        0x718: "device not initialized",
        0x719: "device has a timeout",
        0x71A: "query interface failed",
        0x71B: "wrong interface required",
        0x71C: "class ID is invalid",
        0x71D: "object ID is invalid",
        0x71E: "request is pending",
        0x71F: "request is aborted",
        0x720: "signal warning",
        0x721: "invalid array index",
        0x722: "symbol not active",
        0x723: "access denied",
        0x724: "missing license",
        0x72c: "exception occured during system start",
        0x740: "Error class <client error>",
        0x741: "invalid parameter at service",
        0x742: "polling list is empty",
        0x743: "var connection already in use",
        0x744: "invoke ID in use",
        0x745: "timeout elapsed",
        0x746: "error in win32 subsystem",
        0x747: "Invalid client timeout value",
        0x748: "ads-port not opened",
        0x750: "internal error in ads sync",
        0x751: "hash table overflow",
        0x752: "key not found in hash",
        0x753: "no more symbols in cache",
        0x754: "invalid response received",
        0x755: "sync port is locked",
        0x274c: "A socket operation was attempted to an unreachable host",
}



import ctypes
import struct
import pyads
from .adsresponse import AdsResponse

class ReadResponse(AdsResponse):

    def __init__(self, responseAmsPacket):
        AdsResponse.__init__(self, responseAmsPacket)

        self.Length = struct.unpack_from('I', responseAmsPacket.Data, 4)[0]
        self.Data = responseAmsPacket.Data[8:]


    Length = 0

    Data = b'0'

    def __str__(self):
        result = "AdsReadResponse:\n"
        result += pyads.HexBlock(self.Data)
        return result


    def CreateBuffer(self):
        return ctypes.create_string_buffer(self.Data, len(self.Data))




from .adscommand import AdsCommand
from .deviceinforesponse import DeviceInfoResponse

class DeviceInfoCommand(AdsCommand):

    def __init__(self):
        self.CommandID = 0x0001


    def CreateRequest(self):
        return b''


    def CreateResponse(self, responsePacket):
        return DeviceInfoResponse(responsePacket)


from struct import pack, unpack_from, calcsize

class BinaryParser:

    def __init__(self, byteData = b''):
        self.ByteData = byteData
        self.Position = 0


    ByteData = b''

    Position = 0


    def Append(self, fmt, value):
        self.ByteData = self.ByteData + pack(fmt, value)


    def Unpack(self, fmt):
        result = unpack_from(fmt, self.ByteData, self.Position)
        self.Position = self.Position + calcsize(fmt)

        return result[0]


    def ReadBytes(self, length):
        result = ''
        for i in range(length):
            result = result + chr(self.ReadByte())

        return result


    def WriteBytes(self, byteList):
        if (isinstance(byteList, str)):
            self.ByteData = self.ByteData + byteList
            return

        for b in byteList:
            self.WriteByte(b)




    def ReadUInt8 (self):
        return self.ReadByte()

    def ReadByte(self):
        return self.Unpack('B')

    def WriteUInt8(self, value):
        self.WriteByte(value)

    def WriteByte(self, value):
        self.Append('B', value)




    def ReadInt8(self):
        return self.Unpack('b')

    def WriteInt8(self, value):
        self.Append('b', value)




    def ReadInt16(self):
        return self.Unpack('h')

    def WriteInt16(self, value):
        self.Append('h', value)




    def ReadUInt16(self):
        return self.Unpack('H')

    def WriteUInt16(self, value):
        self.Append('H', value)




    def ReadInt32(self):
        return self.Unpack('i')

    def WriteInt32(self, value):
        self.Append('i', value)




    def ReadUInt32(self):
        return self.Unpack('I')

    def WriteUInt32(self, value):
        self.Append('I', value)




    def ReadInt64(self):
        return self.Unpack('q')

    def WriteInt64(self, value):
        self.Append('q', value)




    def ReadUInt64(self):
        return self.Unpack('Q')

    def WriteUInt64(self, value):
        self.Append('Q', value)




    def ReadDouble(self):
        return self.Unpack('d')

    def WriteDouble(self, value):
        self.Append('d', value)




    def ReadFloat(self):
        return self.Unpack('f')

    def WriteFloat(self, value):
        self.Append('f', value)




import struct
from .adscommand import AdsCommand
from .readwriteresponse import ReadWriteResponse

class ReadWriteCommand(AdsCommand):

    def __init__(self, indexGroup, indexOffset, readLen, dataToWrite = b''):
        self.CommandID = 0x0009
        self.IndexGroup = indexGroup
        self.IndexOffset = indexOffset
        self.ReadLen = readLen
        self.Data = dataToWrite



    IndexGroup = 0

    IndexOffset = 0

    ReadLen = 0

    Data = b''


    def CreateRequest(self):
        result = struct.pack('<II', self.IndexGroup, self.IndexOffset)
        result += struct.pack('<II', self.ReadLen, len(self.Data))
        result += self.Data
        return result


    def CreateResponse(self, responsePacket):
        return ReadWriteResponse(responsePacket)



import struct
from .. import AmsPacket
from .adsresponse import AdsResponse

class WriteResponse(AdsResponse):

    def __init__(self, responseAmsPacket):
        AdsResponse.__init__(self, responseAmsPacket)



import struct
from .adsresponse import AdsResponse

class DeviceInfoResponse(AdsResponse):

    def __init__(self, responseAmsPacket):
        AdsResponse.__init__(self, responseAmsPacket)

        self.MajorVersion = struct.unpack_from('B', responseAmsPacket.Data, 4)[0]
        self.MinorVersion = struct.unpack_from('B', responseAmsPacket.Data, 5)[0]
        self.Build = struct.unpack_from('H', responseAmsPacket.Data, 6)[0]

        deviceNameEnd = 16
        for i in range(8, 24):
            if ord(responseAmsPacket.Data[i]) == 0:
                deviceNameEnd = i
                break

        deviceNameRaw = responseAmsPacket.Data[8:deviceNameEnd]
        self.DeviceName = deviceNameRaw.decode("latin-1").strip(' \t\n\r')


    MajorVersion = 0

    MinorVersion = 0

    Build = 0

    DeviceName = 'Ads Device'


    def __str__(self):
        return "%s (Version %s)" % (self.DeviceName, self.Version())


    def Version(self):
        return "%s.%s.%s" % (self.MajorVersion, self.MinorVersion, self.Build)




#!/usr/bin/env python
"""Pip package for chwiede.pyads
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/chwiede/pyads
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))


# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

from distutils.core import setup

setup(
    name='bfabio.pyads',
    version='0.1.2.1',
    description='python implementation of ADS protocol',
    url='http://www.github.com/chwiede/pyads',

    author='Christoph Wiedemann',
    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3'
    ],

    keywords='beckhoff ads sps',
    packages = find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=[]
)




import struct
from .adscommand import AdsCommand
from .readresponse import ReadResponse

class ReadCommand(AdsCommand):

    def __init__(self, indexGroup, indexOffset, length):
        self.CommandID = 0x0002
        self.IndexGroup = indexGroup
        self.IndexOffset = indexOffset
        self.Length = length



    IndexGroup = 0

    IndexOffset = 0

    Length = 0


    def CreateRequest(self):
        return struct.pack('<III', self.IndexGroup, self.IndexOffset, self.Length)


    def CreateResponse(self, responsePacket):
        return ReadResponse(responsePacket)